#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
A program to download my Lepidoptera data and generate a webpage!

Created on Fri Aug 16 16:23:40 2019
Version : 0.5 beta - 25 Aug 2019
@author: Bhalchandra Pujari 
@email : bspujari@cms.unipune.ac.in



"""




username="bspujari"
location="Ganeshkhind"


#--- MAIN CODE BEGIN NOW
import time,datetime
print ("--- Program Starts "+ datetime.datetime.now().strftime("%d/%B/%Y %H:%M")+"---")
print ("Initializing...")
from halo import Halo #spinnner for loop
import requests
import json
import math,random
import subprocess

import os.path,sys
import collections
from collections import defaultdict
import pandas as pd
from bokeh.plotting import figure
from bokeh.transform import cumsum
from bokeh.embed import autoload_static
from bokeh.layouts import row
from bokeh.resources import CDN
#from scipy.cluster import hierarchy as hc
#import matplotlib.pyplot as plt
#from ete3 import PhyloTree






class iNat_leps():
    
    def main(self):
        
        
        self.name_rank_dict=defaultdict(list)
        #self.phylodf=pd.DataFrame(columns=['Family','Tribe','Genus','Species'])
        self.taxa_id_dump()
        self.iNat_data=[] # empty list to save data
        #DF to create the checklist at the end
        self.checklist_df=pd.DataFrame(columns=['Observation ID','Family','Subfamily','Tribe','Genus','Species'])


        #print ('Downloading the data ... hang on..')
        self.parse_iNat() # fetch the data from iNat and save in above list
        
        #print ("Extracting the data..")
        self.extract_image_data() # extract the relevent data and 
        
    
        print ("Plotting ..")
        self.plots()

        print ("Preparing checklist...")
        self.prepare_checklist()
       
        #print ("Pie charts ...")
        #self.pie_chart()
        
        print ("Creating the webpage..")
        self.write_html("welcome.html")

        
        print ("Everything done. Please check 'welcome.html'. Good bye.\n--- Program Ends "+ datetime.datetime.now().strftime("%d/%B/%Y %H:%M")+"---")
#        print (self.phylodf)
#        z = hc.linkage(self.phylodf.values.T, method='average', metric='correlation')
#        dendrogram = hc.dendrogram(z, labels=self.phylodf.columns)
#        plt.show()
            
#-----MAIN ENDS HERE -----------------------------------


        
    def taxa_id_dump(self): 
        self.taxafile="taxa_id.csv" # a dump file to store taxa-id and names"
        self.TaxaId_Name=pd.DataFrame(columns=("TaxaID", "Name", "RankName", "RankLevel")) # a grand dict storing all IDs and corresponding names. 
        if os.path.isfile(self.taxafile): 
            #print ("Loaiding")
            self.TaxaId_Name =pd.read_csv(self.taxafile, \
                    names=("TaxaID", "Name", "RankName", "RankLevel")\
                    )
        else:
            print("Not exist")
            
        self.TaxaId_Name.astype({'TaxaID': 'int64'})
#        if os.path.isfile(self.taxafile): # read tha database of id-tax available on disk. If nothing is available it will create a database.
#            with open(self.taxafile) as f:
#                for line in f:
#                    (key, val) = line.split()
#                    self.TaxaId_Name[int(key)] = val
                    
    def plots(self): # make plots using bokeh
         plot_items=['family','tribe','genus','species']
         plots=[]
         for item in plot_items :
             pie_data=collections.Counter(self.name_rank_dict[item])
             data=pd.Series(pie_data).reset_index(name='value').rename(columns={'index':item})
             data['angle'] = data['value']/data['value'].sum() * 2*math.pi
             data['color'] = ["#"+''.join([random.choice('0123456789ABCDEF') for j in range(6)]) for i in range(len(data))]
             p = figure(plot_width=275, plot_height=250, title=item.upper()+" ("+str(len(data))+")", toolbar_location=None,
                        tools="hover", tooltips="@"+item+": @value", sizing_mode='scale_width')
             p.wedge(x=0, y=1, radius=0.8,
                     start_angle=cumsum('angle', include_zero=True), end_angle=cumsum('angle'),
                     line_color="white", fill_color='color', 
                     source=data)
             p.axis.visible = False
             p.min_border=0
             plots.append(p)
             
         js_dest="dist/js/bokeh.js"
         js, tag = autoload_static(row(plots), CDN, js_dest)
         f = open(js_dest,"w")
         f.write(js)
         f.close()
         self.htmlplots="\n<center>\n"+tag+"\n</center>\n<br><hr><br>\n\n"

       

    #def nested_pie(self,df):
#
#        df=df.fillna('NA')
#
#        cols = df.columns.tolist()
#        outd = {}
#        gb = df.groupby(cols[0], sort=False).sum()
#        outd[0] = {'names':gb.index.values, 'values':gb.values}
#        for lev in range(1,len(cols)):
#            #print ("lev",df.groupby(cols[:(lev+1)]).sum())
#            gb = df.groupby(cols[:(lev+1)], sort=False).sum()
#            print (gb.index.levels)
#            outd[lev] = {'names':gb.index.levels[lev][gb.index.names[lev]].tolist(),
#                     'values':gb.values}
#        return outd


    def write_html(self,fwrite):
        f= open(fwrite,"w")
        header=open("header.tx","r")
        footer=open("footer.tx","r")
        f.write(header.read())
        f.write(self.htmlplots)
        id_percent=math.ceil((len(self.name_rank_dict['family'])/self.total_observations)*100)
        bar_chart_string="<div class=\"mainbox\">\n<div class='charts'>\n\
        <div class='chart chart--dev'>\
        <span class='chart__title'>Identification status</span>\
        <ul class='chart--horiz'>\
        <li class='chart__bar' style='width:"+str(id_percent)+"%;'>\
          <span class='chart__label'>\
            Identified taxa: "+str(len(self.name_rank_dict['family']))+" of "+str(self.total_observations)+"\
          </span>\
        </li>\
        </ul> \
        </div> \
        </div>"
        
        f.write(bar_chart_string)
        
        #f.write("\n<h3> Click the rank below to see details: </h3>")
        plot_items=['family','tribe','genus','species']
        for item in plot_items :
            f.write("<div class =\"childdiv\">")
            f.write("<button class=\"collapsible\">"+item.upper()+"</button>")
            f.write("<div class=\"content\">")
            f.write("<p>")
            string=" <br>".join(sorted(set(self.name_rank_dict[item])))
            f.write(string)
            f.write("</p>")        
            f.write("</div>")
            f.write("</div>")
            
        f.write("</div>")
        f.write("<div style=\"background-color: #00CC99;color:black;padding:2em\">")
        f.write("<h3> <a href=\"SPPU_Moth_checklist.pdf\"> Download the Checklist in PDF </a></h3>")
        f.write("</div> <hr><br>")
        f.write("<h3> Gallery </h3>")
        f.write(self.html_image_gallery)
        now = datetime.datetime.now().strftime("%d/%B/%Y %H:%M")
        f.write(footer.read().replace('UNKNOWN',now))
        
        f.close()
        footer.close()
        
        
        
    def extract_image_data(self):
        self.html_image_gallery="" # text to be appended in html file
        uniq_sci_name=[]
        spinner = Halo( text_color='blue',spinner='dots3')
        f=open(self.taxafile,'a')  # keep open the taxa_name file to append the unknonwn taxa_name pair
        #rf=open(self.rankdump,'a') # file fo dump rank and rank levels
        self.total_observations=0
        spinner.start(text="Updating the database (takes looong at the first instance)")
        for entry in self.iNat_data : #[:10]:
            
            for i in range(len(entry['results'])):
                if entry['results'][i]['place_guess'] and  location in entry['results'][i]['place_guess']: # Extract  data only from within SPPU
                    ancestry=[int(z) for z in entry['results'][i]['taxon']["min_species_ancestry"].split(",")]
                    if 47157 in ancestry and 47224 not in ancestry : # GET only Lepidoptera (taxa id=47157) excluding butterflies (47224)
                        self.total_observations+=1
                        obs_id=entry['results'][i]['id'] 
                        obs_url="https://www.inaturalist.org/observations/"+str(obs_id)
                        sci_name=entry['results'][i]['taxon']['name'] 
                        rank=entry['results'][i]['taxon']['rank'] # it is a string corresponding to rank_level ('genus'/'family' etc)
                        rank_level=entry['results'][i]['taxon']['rank_level']
                        # ^^^^^^^ what is the ID rank? Is is species or genus or family?
                        taxa_id=entry['results'][i]['taxon']['min_species_taxon_id'] # taxa ID. 
                        
                        thumb_url=entry['results'][i]['observation_photos'][0]['photo']['url'].split('?')[0]
                        original_url=thumb_url.replace("square","original")
                        
                        if rank_level == 10:
                            if sci_name not in uniq_sci_name: # gallery for uniq 'species'only.. but accept multiple for higher ranks
                                uniq_sci_name.append(sci_name)
                                #print (sci_name, rank,rank_level)
                                self.html_image_gallery = self.html_image_gallery +"<a class=\"tooltip\" href=\""+original_url+\
                                   "\" data-lumenbox=\"my-gallery\"  data-title=\""+sci_name+" ("+rank+") <a href=\'"+obs_url+"\' target=\'new\'> Click for details </a> \"> <img src=\""+thumb_url+\
                                    "\" alt=\""+sci_name+"\" /> <span class='tooltiptext'> "+sci_name+" </span>  </a>\n"
                        else:
                            self.html_image_gallery = self.html_image_gallery +"<a class=\"tooltip\" href=\""+original_url+\
                               "\" data-lumenbox=\"my-gallery\"  data-title=\""+sci_name+" ("+rank+") <a href=\'"+obs_url+"\' target=\'new\'> Click for details </a>\" > <img src=\""+thumb_url+\
                                "\" alt=\""+sci_name+"\" /> <span class='tooltiptext'> "+sci_name+" </span> </a>\n"
                                
                        #create a dict of ranks.. not uniq .. species/genus/family etc.. we'll plot this for abundance 
                        #print (self.TaxaId_Name)
                        #print(self.TaxaId_Name.describe())
                        #if not self.TaxaId_Name['TaxaID'].str.contains(str(taxa_id)).any():
                        #    newrow={"TaxaID":str(taxa_id), "Name":sci_name, "RankName":rank, "RankLevel":rank_level}
                        #    self.TaxaId_Name=self.TaxaId_Name.append(newrow, ignore_index=True)
                        # do a thourough search of taxa for name of all the ancesters (like tibe, family etc)

                        #checklist_species=None
                        #checklist_tribe=None
                        #checklist_genus=None
                        #checklist_subfamily=None
                        #checklist_family=None

                        self.checklist_df=self.checklist_df.append({'Observation ID':int(obs_id)},ignore_index=True)


                        for ancester in ancestry :
                            # This loop of the program is terrible. For every
                            # new entry it sends a query to iNat server. May be
                            # there is better way to dump all the data for taxa
                            # and taxaid beforehand? Keeping it as it is, but
                            # needs to revamp.
                            
                            if len(self.TaxaId_Name[self.TaxaId_Name["TaxaID"]==ancester]) == 0: # means no entry for ancester is found in database
                                api_string="http://api.inaturalist.org/v1/taxa/"+str(ancester)
                                r=requests.get(api_string) # download .. make sure proxies are fine
                                try:
                                    r=json.loads(r.text)
                                except :
                                    print ("Error:", ancester, r)
                                taxaname=r['results'][0]['name']                                    
                                rank=r['results'][0]['rank']
                                rank_level=r['results'][0]['rank_level']
                                newrow={"TaxaID":ancester, "Name":taxaname, "RankName":rank, "RankLevel":rank_level}
                                self.TaxaId_Name=self.TaxaId_Name.append(newrow, ignore_index=True)                                
                                f.write(str(ancester)+","+taxaname+","+str(rank) +"," + str(rank_level)+"\n")
                            else:
                                rank=self.TaxaId_Name[self.TaxaId_Name['TaxaID']==ancester]['RankName'].values[0]
                                rank_level=self.TaxaId_Name[self.TaxaId_Name['TaxaID']==ancester]['RankLevel'].values[0]
                                taxaname=self.TaxaId_Name[self.TaxaId_Name['TaxaID']==ancester]['Name'].values[0]
                            
                            #if rank =='family':
                            #    checklist_family=taxaname

                            self.name_rank_dict[rank].append(taxaname)       


                            # update checklist df
                            if rank.title() in self.checklist_df.columns : 
                                self.checklist_df.loc[self.checklist_df['Observation ID']==int(obs_id),rank.title()]=taxaname
                            
                            # FOLLOWING IS FOR THE PHYLO-DATAFRAME 
#                            if rank_level == 30 : #family
#                                current_family=taxaname
#                                newrow={'Family':current_family,'Tribe':'','Genus':'','Species':''}
#                                self.phylodf=self.phylodf.append(newrow,ignore_index=True)
#                            elif rank_level == 25 : # tribe
#                                newrow={'Family':current_family,'Tribe':taxaname,'Genus':'','Species':''}
#                                self.phylodf=self.phylodf.append(newrow,ignore_index=True)
#                                current_tribe=taxaname
#                            elif rank_level == 20 : # genus 
#                                newrow={'Family':current_family,'Tribe':current_tribe,'Genus':taxaname,'Species':''}
#                                self.phylodf=self.phylodf.append(newrow,ignore_index=True)
#                                current_genus=taxaname
#                            elif rank_level == 10 : # species
#                                newrow={'Family':current_family,'Tribe':current_tribe,'Genus':current_genus,'Species':taxaname}
#                                self.phylodf=self.phylodf.append(newrow,ignore_index=True)
                                
                        #if sci_name not in self.name_rank_dict[rank]:
                        
                        
        f.close()
        #rf.close()
        spinner.succeed("Extraction done")
        #print (self.TaxaId_Name)

#    def pie_chart(self):
#
#         outd = self.nested_pie(self.checklist_df[['Family','Subfamily','Genus','Species']])
#         diff = 1/7.0
#         
#         # This first pie chart fill the plot, it's the lowest level
#         plt.pie(outd[6]['values'], labels=outd[6]['names'], labeldistance=0.9,
#                 colors=plt.style.library['bmh']['axes.color_cycle'])
#         ax = plt.gca()
#         # For each successive plot, change the max radius so that they overlay
#         for i in np.arange(5,-1,-1):
#             ax.pie(outd[i]['values'], labels=outd[i]['names'],
#                    radius=np.float(i+1)/7.0, labeldistance=((2*(i+1)-1)/14.0)/((i+1)/7.0),
#                    colors=plt.style.library['bmh']['axes.color_cycle'])
#         ax.set_aspect('equal')
#         plt.savefig("Nested.png")

    def prepare_checklist(self):
        self.checklist_df['Observation ID']=self.checklist_df['Observation ID'].astype('int')  
        self.checklist_df.dropna(subset=['Family'],inplace=True)
        self.checklist_df=self.checklist_df.sort_values(by=self.checklist_df.columns.to_list()[1:]) # sort but skip 'Observation ID' column
        self.checklist_df=self.checklist_df.drop_duplicates(subset=self.checklist_df.columns.to_list()[1:])
        self.checklist_df=self.checklist_df.reset_index()
        self.checklist_df.index += 1 
        f=open("checklist.tex",'w')
        f.write(self.checklist_df.to_latex(
            columns=['Family','Subfamily','Genus','Species','Observation ID'], 
            header=['Family','Subfamily','Genus','Species','Obs ID'],
            caption="List of moths of Savitribai Phule Pune University, retrieved on "+datetime.datetime.now().strftime("%b %d, %Y."),
            longtable=True,na_rep=' '))
        f.close()
        print ("Compiling the checklist. You need functioning LaTeX environment\n")
        subprocess.run(["pdflatex","SPPU_Moth_checklist",">","/dev/null"])
        print ("Please check SPPU_Moth_checklist.pdf file for the checklist!")
        #f=open("checklist.tex",'w')
        #f.write("\\documentclass[12pt]{article}\n\\usepackage{booktabs,longtable,lscape}\n\
        #        \\title{Checklist for Moths at Savitribai Phule Pune University} \
        #        \\author{Bhalchandra Pujari} \\begin{document} \\maketitle \
        #        {\\tiny\
        #        ")
        #f.write(self.checklist_df.to_latex(columns=['Family','Subfamily','Genus','Species','Observation ID'],longtable=True,na_rep=' '))
        #f.write("\n }\n \
        #        Note: This file is automatically generated by fetching the data\
        #        from iNaturalist.org website. You can check the individual\
        #        species by using the Observation ID column: \\\\ \\verb+ inaturalist.org/observations/<<Observation ID>> +\\  It is also\
        #        likely that the identification may change over the period of\
        #        time. It is advisable to update this checklist to its latest\
        #        version by visiting \\\\ \\verb+ http://cms.unipune.ac.in/~bspujari/Pune_university_Lepidoptera/+.\
        #        \\\\ For more details write to \\verb+ bspujari @ scms.unipune.ac.in+ \\end{document}")
#
#        f.close()
#        print ("checklist.tex file is ready for further processing\n")
        self.checklist_df.to_csv("checklist.csv")


            

    def parse_iNat(self): 
      
      page_count=1 # Required by the format of API 
           # total observations = pages * observations per page
           #                  (set to 100 at this stage)
      total_pages='NA'
      
      
      
      spinner = Halo( text_color='green',spinner='circleQuarters')
      t0=time.time()
      while True:
      #while page_count <2:
          
          if page_count == 1 : 
              spin_text="Downloading page "+str(page_count)
          else :
              spin_text="Downloading page "+str(page_count)+" of "+ total_pages+\
              "  Time Remaining : " + str(math.ceil((t1-t0)*(int(total_pages)-page_count+1))) + " sec"
          spinner.start(text=spin_text )
          

              # Build the string for API download
          api_string="http://api.inaturalist.org/v1/observations?"+\
          "user_id="+username+\
          "&identified=true"+\
          "&photos=true"+\
          "&page="+str(page_count)+\
          "&per_page=100"+\
          "&order=desc"+\
          "&order_by=created_at"
          # looks like 100 is the limit for per page observations.
          
          try:
              r=requests.get(api_string) # download .. make sure proxies are fine
          except :
              print ("\nERROR: Active internet connection is required to download the data. Please check you connection and try again.\nI have to stop here. Bye.")
              sys.exit()
          if not r : 
              print ("\nERROR! Did not receive any data. Please check your network connectivity. Are you behind firewall/proxy?")
              sys.exit()
              
          j=json.loads(r.text)
          
          
          if page_count == 1 : # for the first iteration, determine how many pages of downloadng is there..
              total_pages=str(math.ceil(j['total_results']/100) +1) # total results/results per page
              t1=time.time()
              
          if j['results'] : # error will occur when page numbers are exhausted.. till that time keep increasing page numbers.
              page_count+=1
              self.iNat_data.append(json.loads(r.text))
          else:
              break
          
          
          spinner.stop()
      spinner.succeed("Download complete") 
if __name__ == '__main__':
   u=iNat_leps()
   u.main()
